using System;
using System.Net;
using System.IO;
using System.Collections.Specialized;

namespace TwitterApi
{
    public class oAuthTwitter : OAuthBase
    {

        private string _consumerKey = "";
        private string _consumerSecret = "";

        private string _token = "";
        private string _tokenSecret = "";
        public enum Method { GET, POST };
        public const string RequestTokenUrl = "http://twitter.com/oauth/request_token";
        public const string AuthorizeUrl = "http://twitter.com/oauth/authorize";
        public const string AccessTokenUrl = "http://twitter.com/oauth/access_token";
        #region Properties
        /// <summary>
        /// The Consumer Key of your app goes here
        /// </summary>
        public string ConsumerKey 
        {
            get
            {
                return _consumerKey; 
            } 
            set { _consumerKey = value; } 
        }

        /// <summary>
        /// The Consumer Secret of your app goes here
        /// </summary>
        public string ConsumerSecret { 
            get {
                return _consumerSecret; 
            } 
            set { _consumerSecret = value; } 
        }

        public string Token { get { return _token; } set { _token = value; } }
        public string TokenSecret { get { return _tokenSecret; } set { _tokenSecret = value; } }
#endregion

        /// <summary>
        /// Get the link to Twitter's authorization page for this application.
        /// </summary>
        /// <returns>The url with a valid request token, or a null string.</returns>
        public string AuthorizationLinkGet()
        {
            string dummy;
            return AuthorizationLinkGet(out dummy);
        }

        public string AuthorizationLinkGet(out string authToken)
        {
            string ret = null;
            authToken = null;

            string response = oAuthWebRequest(Method.GET, RequestTokenUrl, String.Empty);
            if (response.Length > 0)
            {
                //response contains token and token secret.  We only need the token.
                NameValueCollection qs = StringUtil.ParseQueryString(response);
                if (qs["oauth_token"] != null)
                {
                    authToken = qs["oauth_token"];
                    ret = AuthorizeUrl + "?oauth_token=" + authToken;
                }
            }
            return ret;
        }

        /// <summary>
        /// Exchange the request token for an access token.
        /// </summary>
        /// <param name="requestToken">The oauth_token is supplied by Twitter's authorization page following the callback.</param>
        public void AccessTokenGet(string requestToken)
        {
            this.Token = requestToken;

            string response = oAuthWebRequest(Method.GET, AccessTokenUrl, String.Empty);

            if (response.Length > 0)
            {
                //Store the Token and Token Secret
                NameValueCollection qs = StringUtil.ParseQueryString(response);
                if (qs["oauth_token"] != null)
                {
                    this.Token = qs["oauth_token"];
                }
                if (qs["oauth_token_secret"] != null)
                {
                    this.TokenSecret = qs["oauth_token_secret"];
                }
            }
        }

        /// <summary>
        /// Exchange the request token for an access token using a pin from Twitter.
        /// </summary>
        /// <param name="pin">Pin code entered by user</param>
        /// <param name="authToken">The oauth_token is supplied by Twitter's authorization page following the callback.</param>
        public void AccessTokenGet(string authToken, string pin )
        {
            this.Token = authToken;

            string response = oAuthWebRequest(Method.GET, AccessTokenUrl, null, pin);

            if (response.Length > 0)
            {
                //Store the Token and Token Secret
                NameValueCollection qs = StringUtil.ParseQueryString(response);
                if (qs["oauth_token"] != null)
                {
                    this.Token = qs["oauth_token"];
                }
                if (qs["oauth_token_secret"] != null)
                {
                    this.TokenSecret = qs["oauth_token_secret"];
                }
            }
        }
        
        /// <summary>
        /// Submit a web request using oAuth.
        /// </summary>
        /// <param name="method">GET or POST</param>
        /// <param name="url">The full url, including the querystring.</param>
        /// <param name="postData">Data to post (querystring format)</param>
        /// <returns>The web server response.</returns>
        public string oAuthWebRequest(Method method, string url, string postData)
        {
            return oAuthWebRequest(method, url, postData, null);
        }

        /// <summary>
        /// Submit a web request using oAuth.
        /// </summary>
        /// <param name="method">GET or POST</param>
        /// <param name="url">The full url, including the querystring.</param>
        /// <param name="postData">Data to post (querystring format)</param>
        /// <param name="verifier">Verifier (pin) or null if irellevant</param>
        /// <returns>The web server response.</returns>
        public string oAuthWebRequest(Method method, string url, string postData, string verifier)
        {
            string outUrl = "";
            string querystring = "";
            string ret = "";

            //Setup postData for signing.
            //Add the postData to the querystring.
            if (method == Method.POST)
            {
                if (postData.Length > 0)
                {
                    //Decode the parameters and re-encode using the oAuth UrlEncode method.
                    NameValueCollection qs = StringUtil.ParseQueryString(postData);
                    postData = "";
                    foreach (string key in qs.AllKeys)
                    {
                        if (postData.Length > 0)
                        {
                            postData += "&";
                        }
                        qs[key] = StringUtil.UrlDecode(qs[key]);
                        qs[key] = StringUtil.UrlEncode(qs[key]);
                        postData += key + "=" + qs[key];

                    }
                    if (url.IndexOf("?") > 0)
                    {
                        url += "&";
                    }
                    else
                    {
                        url += "?";
                    }
                    url += postData;
                }
            }

            Uri uri = new Uri(url);
            
            string nonce = this.GenerateNonce();
            string timeStamp = this.GenerateTimeStamp();

            //Generate Signature
            string sig = this.GenerateSignature(uri,
                this.ConsumerKey,
                this.ConsumerSecret,
                this.Token,
                this.TokenSecret,
                verifier,
                method.ToString(),
                timeStamp,
                nonce,
                out outUrl,
                out querystring);

            querystring += "&oauth_signature=" + StringUtil.UrlEncode(sig);

            //Convert the querystring to postData
            if (method == Method.POST)
            {
                postData = querystring;
                querystring = "";
            }

            if (querystring.Length > 0)
            {
                outUrl += "?";
            }

            ret = WebRequest(method, outUrl +  querystring, postData);

            return ret;
        }

        /// <summary>
        /// Web Request Wrapper
        /// </summary>
        /// <param name="method">Http Method</param>
        /// <param name="url">Full url to the web resource</param>
        /// <param name="postData">Data to post in querystring format</param>
        /// <returns>The web server response.</returns>
        public static string WebRequest(Method method, string url, string postData)
        {
            HttpWebRequest webRequest = null;
            StreamWriter requestWriter = null;
            string responseData = "";

            webRequest = System.Net.WebRequest.Create(url) as HttpWebRequest;
            webRequest.Method = method.ToString();

            if (method == Method.POST)
            {
                webRequest.ContentType = "application/x-www-form-urlencoded";
                webRequest.ContentLength = postData.Length;

                //POST the data.
                requestWriter = new StreamWriter(webRequest.GetRequestStream());
                try
                {
                    requestWriter.Write(postData);
                }
                finally
                {
                    requestWriter.Close();
                }
            }
            responseData = WebResponseGet(webRequest);
            return responseData;

        }

        /// <summary>
        /// Process the web response.
        /// </summary>
        /// <param name="webRequest">The request object.</param>
        /// <returns>The response data.</returns>
        public static string WebResponseGet(HttpWebRequest webRequest)
        {
            StreamReader responseReader = null;
            string responseData = "";
            var theStream = webRequest.GetResponse().GetResponseStream();

            try
            {
                responseReader = new StreamReader(theStream);
                responseData = responseReader.ReadToEnd();
                
            }
            finally
            {
                theStream.Close();
                responseReader = null;
            }

            return responseData;
        }
    }
}
